<?php
/**
 * Abstract class for handling Metaboxes in Wordpress
 */
namespace EwTools\Admin;

use EwTools\Models\Page;
use EwTools\Support\Arr;
use EwTools\Support\Str;
use EwTools\Twig\View;
use \WP_Post;

//use WWW\Classes\Customer\EWC_Config;

/**
 * Metabox
 *
 * @link    http://infomaker.se
 * @package EwTools\Admin
 * @since   EwTools\Admin\Metabox 0.1
 */
abstract class Metabox {
    
    /**
     * Contains the id/key where the data will be saved in Wordpress database
     *
     * @var string
     */
    public static $postmeta_id = 'ew_meta_box';
    
    /**
     * Contains the context within the screen where the boxes should display.
     * Page edit screen contexts include 'normal', 'side' and 'advanced'. Default 'side'.
     *
     * @var string
     */
    protected static $box_position = 'side';
    
    /**
     * Contains the priority within the context where the boxes
     * should show ('high', 'low', 'core', 'sorted', 'default'). Default 'default'.
     *
     * @var string
     */
    protected static $box_priority = 'default';
    
    /**
     * If set, will contain the path to the page-template that will have to be chosen for the metabox to be visible
     *
     * @var string
     */
    public static $required_template = null;
    
    /**
     * Contains the id attribute of the html-metabox and is the default value for "meta_input_prefix"
     *
     * @var string
     */
    protected $meta_box_id = 'ew-meta-box';
    
    /**
     * Contains the nonce name used to identify meta changes
     *
     * @var string
     */
    protected $meta_box_nonce = 'ew_meta_nonce';
    
    /**
     * Contains the text domain for the metabox. Defaults to everywares textdomain for metaboxes
     *
     * @var string
     */
    protected $text_domain = 'ewtools_textdomain';
    
    /**
     * Contains the prefix for name and id attributes on form inputs
     *
     * @var null|string
     */
    protected $meta_input_prefix = null;
    
    /**
     * Contains the fields of the metabox form.
     *
     * @var array
     */
    protected $fields = [];
    
    /**
     * Contains a list of rules for the form fields used for validation
     *
     * @var array
     */
    protected $rules = [];
    
    /**
     * Contains the auto generated view data
     *
     * @var array
     */
    protected $view_data = [];
    
    /**
     * Contains the path to where the templates are stored
     *
     * @var string
     */
    protected $template_path = '@tools/widgets/';
    
    /**
     * Contains the Metabox untranslated name
     *
     * @var string
     */
    protected $name = 'Metabox Name';
    
    /**
     * Contains the Metabox untranslated description
     *
     * @var string
     */
    protected $description = 'Metabox Description';
    
    /**
     * Contains the Metabox version
     *
     * @var string
     */
    protected $version = '0.1';
    
    /**
     * Contains the version a metabox will get if no version exist
     *
     * @var string
     */
    private $default_version = '0.1';
    
    /**
     * Contains the type of post that the metabox will be bound to
     *
     * @var string
     */
    protected $post_type = 'page';
    
    /**
     * Metabox constructor.
     *
     * @uses  $pagenow
     * @uses  is_admin()
     * @uses  add_action('add_meta_boxes')
     * @uses  add_action('save_post')
     * @uses  Str::snake()
     *
     * @param $headline
     *
     * @since 0.1
     */
    public function __construct( $headline = '' ) {
        if( Str::notEmpty( $headline ) ) {
            $this->name = $headline;
        }
        
        // Default prefix to to box id if not set.
        if( is_null( $this->meta_input_prefix ) ) {
            $this->meta_input_prefix = $this->meta_box_id;
        }
        
        $this->meta_box_nonce = Str::snake( $this->meta_input_prefix . "-nonce" );
        
        if ( is_admin() ) {
            add_action( 'load-post.php',     array( $this, 'init' ) );
            add_action( 'load-post-new.php', array( $this, 'init' ) );
        }
    }
    
    // Public Functions
    // ======================================================
    
    public function init() {
        add_action( 'add_meta_boxes', [ &$this, 'setupMetabox' ] );
        add_action( 'save_post', [ &$this, 'saveForm' ], 10, 3 );
    }
    
    /**
     * Initialize the metabox form
     *
     * @param WP_Post $post
     *
     * @since 0.1
     * @return void
     */
    public function initializeForm( WP_Post $post ) {
        $this->form( ( new Page( $post ) )->getMeta( static::$postmeta_id, [] ) );
    }
    
    /**
     * Fires on "save_post"
     *
     * @param int $post_id The post ID.
     * @param WP_Post $post The post object.
     * @param bool $update Whether this is an existing post being updated or not.
     *
     * @since 0.1
     * @return string
     */
    public function saveForm( $post_id, WP_Post $post, $update ) {
        
        // check if there was a multisite switch before
        if ( is_multisite() && ms_is_switched() ) {
            return $post_id;
        }
        
        $page = new Page( $post );
        
        if( $this->shouldUpdate($post_id) && $page->updateAllowed() ) {
            $meta_data = $this->update( $_POST[ $this->meta_input_prefix ] ?: [], $page->getMeta( static::$postmeta_id, [] ) );
            
            if ( $meta_data ) {
                $page->updateMeta( static::$postmeta_id, $meta_data );
            }
        }
    }
    
    /**
     * Function for registering the metabox in Wordpress
     *
     * @uses  add_meta_box()
     * @since 0.1
     * @return void
     */
    public function setupMetabox() {
        if( $this->hasRequiredTemplate() ) {
            add_meta_box(
                $this->meta_box_id,
                $this->translate( $this->name ),
                [ &$this, 'initializeForm' ],
                'Page',
                static::$box_position,
                static::$box_priority
            );
        }
    }
    
    /**
     * Render the form wrapper. This is the for that calls the form-method
     *
     * @param array $instance
     *
     * @since 0.1
     * @return void
     */
    public function form( $instance ) {
        
        $is_new_instance = empty( $instance );
    
        // Add current widget version to new instances of the widget
        if( $is_new_instance ) {
            $instance[ 'metabox_version' ] = $this->version;
        }
        
        // Add default fields on new widgets
        $this->fields = array_replace_recursive( [
            'metabox_version' => $this->default_version
        ], $this->fields );
        
        $instance = array_replace_recursive( $this->fields, $instance );
        
        $this->view_data = [
            'metabox' => [
                'name'         => $this->name,
                'description'  => $this->description,
                'class_prefix' => $this->getFieldId( 'form' ),
                'id_prefix'    => $this->meta_box_id,
                'version'      => $this->version,
                'up_to_date'   => $this->isUpToDate( $instance ),
                'text_domain'  => $this->text_domain
            ]
        ];
        
        //update version on instance to be updated on save
        $instance[ 'metabox_version' ] = $this->version;
        
        $template_array = array_replace_recursive( $this->view_data, $this->setupFields( $instance ) );
        
        // Send the instance to the correct function
        if( $is_new_instance ) {
            $template_array = $this->onCreate( $template_array );
        }
        
        $this->edit( $template_array );
        
        wp_nonce_field( $this->meta_box_id, $this->meta_box_nonce );
    }
    
    /**
     * When the post is saved, saves our custom data.
     *
     * @param array $new_instance Values just sent to be saved.
     * @param array $old_instance Previously saved values from database.
     *
     * @since 0.1
     * @return array
     */
    public function update( $new_instance, $old_instance ) {
        
        // Update the form if the new data is valid
        if( $this->validate( $new_instance, $this->rules ) ) {
            
            $new_instance = array_replace_recursive( $this->fields, $new_instance );
            
            // Store if new form or change existing form
            if( empty( $old_instance ) ) {
                return $this->onStore( $new_instance );
            } else {
                return $this->onChange( $new_instance, $old_instance );
            }
        }
        
        return $old_instance;
    }
    
    // Protected functions
    // ======================================================
    
    /**
     * Create a field-id used on form-input
     *
     * @param string $field_name
     *
     * @since 0.1
     * @return string
     */
    protected function getFieldId( $field_name ) {
        return $this->meta_input_prefix . '-' . trim( str_replace( [ '[]', '[', ']' ], [
                '',
                '-',
                ''
            ], $field_name ), '-' );
    }
    
    /**
     * Return an array with the id and name from a multi valued field for the form input
     *
     * @param $field
     * @param $instance
     *
     * @since 0.1
     * @return array
     */
    protected function getFieldMultiValues( $field, $instance ) {
        $return_val = [];
        foreach ( $instance as $i => $value ) {
            
            if( ! is_array( $value ) ) {
                return $this->getFieldValues( "{$field}[]", $instance );
            }
            
            foreach ( $value as $settings_name => $settings_value ) {
                $return_val[ $i ][ $settings_name ] = $this->getFieldValues( "{$field}[{$i}][{$settings_name}]", $instance[ $i ][ $settings_name ] );
            }
        }
        
        return $return_val;
    }
    
    /**
     * Create a field-name used on form-input
     *
     * @param string $field_name
     *
     * @since 0.1
     * @return string
     */
    protected function getFieldName( $field_name ) {
        if( false === $pos = strpos( $field_name, '[' ) ) {
            return $this->meta_input_prefix . "[{$field_name}]";
        }
        
        return $this->meta_input_prefix . '[' . substr_replace( $field_name, '][', $pos, strlen( '[' ) );
    }
    
    /**
     * Return an array with the id and name for the form input
     *
     * @param $field
     * @param $value
     *
     * @since 0.1
     * @return array
     */
    protected function getFieldValues( $field, $value ) {
        
        return [
            'id'    => $this->getFieldId( $field ),
            'name'  => $this->getFieldName( $field ),
            'value' => $value
        ];
    }
    
    /**
     * Retrieve the name of the required template
     *
     * @since 0.1
     * @return string
     */
    protected function getTemplateName() {
        if( static::$required_template ) {
            $template_file = Str::startsWith( static::$required_template, '/' ) ? static::$required_template : '/' . static::$required_template;
            $located_file = locate_template( $template_file );
            
            if( !empty($located_file) ) {
                $file_data = array_replace( [ 'name' => $located_file ], get_file_data( $located_file, [ 'name' => 'Template Name' ] ) );
                
                return $file_data[ 'name' ];
            }
        }
        
        return '';
    }
    
    /**
     * Check if the required page-template is set to the page
     *
     * @since 0.1
     * @return bool
     */
    protected function hasRequiredTemplate() {
        if( static::$required_template ) {
            return static::$required_template === get_page_template_slug( get_queried_object_id() );
        }
        
        return true;
    }
    
    /**
     * Fires on saving updated metabox
     *
     * @param array $new_instance
     * @param array $old_instance
     *
     * @since 0.1
     * @return array $new_instance The form data to be saved
     */
    protected function onChange( $new_instance, $old_instance ) {
        return $new_instance;
    }
    
    /**
     * Fires on creating new metabox
     *
     * @param array $instance
     *
     * @since 0.1
     * @return array The metabox settings
     */
    protected function onCreate( $instance ) {
        return $instance;
    }
    
    /**
     * Fires on saving new metabox
     *
     * @param array $new_instance
     *
     * @since 0.1
     * @return array $new_instance The form data to be saved
     */
    protected function onStore( $new_instance ) {
        return $new_instance;
    }
    
    /**
     * Function to verify this came from our screen and with proper authorization,
     * because the save_post action can be triggered at other times.
     *
     * @param string $post_id
     *
     * @since 0.1
     * @return bool
     */
    protected function shouldUpdate($post_id) {
        
        // Check if not an autosave.
        if ( wp_is_post_autosave( $post_id ) ) {
            return false;
        }

        // Check if not a revision.
        if ( wp_is_post_revision( $post_id ) ) {
            return false;
        }
        
        // Verify that the nonce is valid.
        return isset( $_POST[ $this->meta_box_nonce ] ) ? wp_verify_nonce( $_POST[ $this->meta_box_nonce ], $this->meta_box_id ) : false;
    }
    
    /**
     * Function to setup field-values for the view
     *
     * @param array $instance
     *
     * @since 0.1
     * @return array
     */
    protected function setupFields( $instance ) {
        $fields_values = [];
        
        foreach ( $this->fields as $field => $value ) {
            $fields_values[ $field ] = is_array( $value ) ? $this->getFieldMultiValues( $field, $instance[ $field ] ) : $this->getFieldValues( $field, $instance[ $field ] );
        }
        
        return $fields_values;
    }
    
    /**
     * Function for validating the form
     *
     * @param array $new_instance
     * @param array $rules
     *
     * @since 0.1
     * @return bool
     */
    protected function validate( array $new_instance, array $rules ) {
        return true;
    }
    
    /**
     * Function to render a template from the set path
     *
     * @param string $template
     * @param array  $template_array
     *
     * @since 0.1
     * @return void
     */
    protected function view( $template = 'form', $template_array ) {
        View::render( Str::finish( $this->template_path, '/' ) . $template, $template_array );
    }
    
    /**
     * Extract the data from the Metabox dock-header
     * and save into properties to be used when instantiating the widget
     *
     * @uses  get_file_data()
     *
     * @param string $file The metabox file with the docBlock
     *
     * @since 0.1
     * @return array the extracted data
     */
    protected function setMetaboxData( $file ) {
        
        $widget_data = get_file_data( $file, [
            'name'        => 'Metabox Name',
            'description' => 'Description',
            'version'     => 'Version'
        ] );
        
        $this->name        = $this->translate( $widget_data[ 'name' ] );
        $this->description = $this->translate( $widget_data[ 'description' ] );
        $this->version     = $this->translate( $widget_data[ 'version' ] );
        
        return $widget_data;
    }
    
    /**
     * Function to translate text using the widgets text-domain
     *
     * @uses  __()
     *
     * @param $text
     *
     * @since 0.1
     * @return mixed
     */
    protected function translate( $text ) {
        return __( $text, $this->text_domain );
    }
    
    /**
     * Function required by the metabox-child. Returns the string-representation of the metabox-form
     *
     * @param array $view_data
     *
     * @since 0.1
     * @return void
     */
    abstract protected function edit( $view_data );
    
    /**
     * Function to compare the version of the widget and the saved instance
     *
     * @param $instance
     *
     * @since 0.1
     * @return bool
     */
    private function isUpToDate( $instance ) {
        return version_compare( $instance[ 'metabox_version' ], $this->version ) >= 0;
    }
    
    /**
     * Function for setting up the metabox
     *
     * @param string $headline
     *
     * @since 0.1
     * @return static
     */
    public static function setup( $headline = '' ) {
        return new static( $headline );
    }
    
    /**
     * Retrieve the meta-data used by the metabox.
     * Retrieve all data or use the optional field_name to specify a field
     *
     * @param Page $page
     * @param string  $field_name
     *
     * @since 0.1
     * @return array|mixed
     */
    /**
     * Retrieve the meta-data used by the metabox.
     * Retrieve all data or use the optional field_name to specify a field
     *
     * @param Page $page
     * @param null $field_name
     *
     * @since 0.1
     * @return mixed
     */
    public static function getMetaData( Page $page, $field_name = null ) {
        $meta = $page->getMeta( static::$postmeta_id, [] );
        
        // Return the field if specified
        if( $field_name ) {
            return Arr::get( $meta, $field_name, null );
        }
        
        return $meta;
    }
}