<?php
/**
 * Handler: to create an easy way to build urls for using Imengine API.
 */
namespace EwTools\Handler;

use RuntimeException;

/**
 * Imengine
 *
 * @link    http://infomaker.se
 * @package EwTools\Handler
 * @since   EwTools\Handler\Imengine 0.1
 */
class Imengine {
    
    /**
     * Contains the base url to the Imengine server
     *
     * @var string
     */
    private $base_url;
    
    /**
     * Contains an array of the parameters that will be sent to Imengine server
     *
     * @var array
     */
    private $params;
    
    /**
     * Contains the default parameters that's always sent to Imengine regardless of the function used
     *
     * @var array
     */
    protected $default_params = [
        'function' => null,
        'type'     => 'preview', // preview|primary
        'source'   => 'false',
        'q'        => 75,
    ];
    
    /**
     * Imengine constructor.
     *
     * @since 0.1
     *
     * @param string $imengine_url
     * @param array  $params
     */
    private function __construct( $imengine_url, $params = [] ) {
        $this->base_url = $imengine_url;
        $this->params   = array_replace( $this->default_params, $params );
    }
    
    /**
     * Use an uuid to specify witch image to manipulate.
     * (The image must exist in the Open content configured in the Imengine api in use)
     *
     * @param string $uuid
     *
     * @since 0.1
     * @return null|string
     */
    public function fromUuid( $uuid = '' ) {
        return $this->buildFromParams( array_merge( [ 'uuid' => $uuid ], $this->params ) );
    }
    
    /**
     * Use an external image to manipulate.
     * (The new version will be saved on Imengine server)
     *
     * @param string $imgsrc
     *
     * @since 0.1
     * @return null|string
     */
    public function fromSrc( $imgsrc = '' ) {
        
        // Add format if not already set
        if( ! isset( $this->params[ 'format' ] ) ) {
            $this->params[ 'format' ] = pathinfo( strtok( $imgsrc, '?' ), PATHINFO_EXTENSION );
        }
        
        return $this->buildFromParams( array_merge( [ 'imgsrc' => trim( $imgsrc ) ], $this->params ) );
    }
    
    // Chainable functions: functions to chain on mirror functions to further manipulate your query before sending it to Imengine
    // ======================================================
    
    /**
     * Add or override parameters not specific to the function being used
     *
     * @param array $params
     *
     * @since 0.1
     * @return $this
     */
    public function with( array $params = [] ) {
        $this->params = array_replace( $this->params, $params );
        
        return $this;
    }
    
    /**
     * Function to allow for quality change of the requested image
     *
     * @param int $quality [1-100]
     *
     * @since 0.1
     * @return $this
     */
    public function quality( $quality ) {
        if( is_int( $quality ) && $quality > 0 && $quality <= 100 ) {
            $this->params = array_replace( $this->params, [ 'q' => $quality ] );
        }
        
        return $this;
    }
    
    // Protected functions:
    // ======================================================
    
    /**
     * Function to build the query url to Imengine API
     *
     * @param array $params
     *
     * @since 0.1
     * @return string
     */
    protected function buildFromParams( array $params = [] ) {
        return "{$this->base_url}?" . http_build_query( $params );
    }
    
    /**
     * Function to calculate height from given width using ration
     *
     * @param $width
     * @param $ratio
     *
     * @since 0.1
     * @return int
     */
    protected static function calculateRatioHeight( $width, $ratio ) {
        
        $ratio_arr    = explode( ':', $ratio );
        $ratio_width  = intval( $ratio_arr[ 0 ] );
        $ratio_height = intval( $ratio_arr[ 1 ] );
        
        return intval( round( ( $width / $ratio_width ) * $ratio_height ) );
    }
    
    /**
     * Function to create an instance with only params.
     * Used by all Imengine server-function mirrored functions
     *
     * @param $func
     * @param $params
     *
     * @since 0.1
     * @return static
     */
    protected static function createMirrorFunction( $func, $params = [] ) {
        return new static( static::getImengineUrl(), array_merge( [ 'function' => $func ], $params ) );
    }
    
    /**
     * Function to get DEV or PROD img-base URL (with or without CDN)
     *
     * @since 0.1
     * @return string url
     */
    protected static function getImengineUrl() {
        if( ! defined( 'CF_IMENGINE' ) ) {
            throw new RuntimeException( 'CF_IMENGINE is not defined and can therefor not serve images' );
        }
        
        return defined( 'CF_IMENGINE' ) ? CF_IMENGINE : false;
    }
    
    // Mirror functions: mirrors the functions found on Imengine API
    // ======================================================
    
    /**
     * Scale image to cover the limits.
     * (Can exceed limits)
     *
     * @param int $width
     * @param int $height
     *
     * @since 0.1
     * @return Imengine
     */
    public static function cover( $width = null, $height = null ) {
        return static::createMirrorFunction( 'cover', compact( [ 'function', 'width', 'height' ] ) );
    }
    
    /**
     * Function to create a url from scratch
     *
     * @param array $params
     *
     * @since 0.1
     * @return static
     */
    public static function create( $params = [] ) {
        return new static( static::getImengineUrl(), $params );
    }
    
    /**
     * Crop image into defined crop-size from custom start-positions.
     * Then scale into defined size
     *
     * @param int $crop_w
     * @param int $crop_h
     * @param int $width
     * @param int $height
     * @param int $start_x
     * @param int $start_y
     *
     * @since 0.1
     * @return Imengine
     */
    public static function cropResize( $crop_w, $crop_h, $width = null, $height = null, $start_x = 0, $start_y = 0 ) {
        return static::createMirrorFunction( 'cropresize', compact( [
            'crop_w',
            'crop_h',
            'start_x',
            'start_y',
            'width',
            'height'
        ] ) );
    }
    
    /**
     * Scale image to fit inside the limits.
     * Add scale up to make the image cover the maxsize-area
     * (Can exceed limits on scale up)
     *
     * @param int  $maxsize
     * @param bool $scaleup
     *
     * @since 0.1
     * @return Imengine
     */
    public static function fit( $maxsize, $scaleup = true ) {
        return static::createMirrorFunction( 'fit', compact( [ 'maxsize', 'scaleup' ] ) );
    }
    
    /**
     * Crop image into defined size from center position
     *
     * @param int $width
     * @param int $height
     *
     * @since 0.1
     * @return Imengine
     */
    public static function hardcrop( $width, $height ) {
        return static::createMirrorFunction( 'hardcrop', compact( [ 'width', 'height' ] ) );
    }
    
    /**
     * Crop image into defined size from custom start-positions.
     *
     * @param int $width
     * @param int $height
     * @param int $start_x
     * @param int $start_y
     * @param int $zoom
     *
     * @since 0.1
     * @return Imengine
     */
    public static function npCrop( $width, $height, $start_x = 0, $start_y = 0, $zoom = 1 ) {
        return static::createMirrorFunction( 'np_crop', compact( [
            'width',
            'height',
            'start_x',
            'start_y',
            'zoom'
        ] ) );
    }
    
    /**
     * Fetch image with original size
     *
     * @since 0.1
     * @return Imengine
     */
    public static function original() {
        return static::createMirrorFunction( 'original' );
    }
    
    /**
     * Crop image into defined size from center position
     *
     * @param int $width
     * @param int $height
     *
     * @since 0.1
     * @return Imengine
     */
    public static function thumbnail( $width, $height ) {
        return static::createMirrorFunction( 'thumbnail', compact( [ 'width', 'height' ] ) );
    }
    
    // Extended mirror functions: functions with extended functionality och the basic mirror functions
    // ======================================================
    
    /**
     * Crop image into given ratio
     *
     * @param $width
     * @param $ratio
     *
     * @since 0.1
     * @return Imengine
     */
    public static function ratio( $width, $ratio = '3:2' ) {
        return static::hardcrop( $width, static::calculateRatioHeight( $width, $ratio ) );
    }
    
    /**
     * Function for fetching the Raw format of the image (May have an unlimited size)
     *
     * @param string $uuid
     * @param int    $quality
     *
     * @since 0.1
     * @return void
     */
    public static function raw( $uuid, $quality = 100 ) {
        return ( new static( static::getImengineUrl(), [ 'type' => 'primary', 'q' => $quality ] ) )->fromUuid( $uuid );
    }
    
    /**
     * Crop image into a square
     *
     * @param $size
     *
     * @since 0.1
     * @return Imengine
     */
    public static function square( $size ) {
        return static::hardcrop( $size, $size );
    }
}
