<?php
/**
 * Support: for advanced string handling
 */
namespace EwTools\Support;

use EwTools\Support\Traits\Macroable;

/**
 * Str
 *
 * @link    http://infomaker.se
 * @package EwTools\Support
 * @since   EwTools\Support\Str 0.1
 */
class Str {
    
    use Macroable;
    
    /**
     * The cache of snake-cased words.
     *
     * @var array
     */
    protected static $snakeCache = [];
    
    /**
     * The cache of camel-cased words.
     *
     * @var array
     */
    protected static $camelCache = [];
    
    /**
     * The cache of studly-cased words.
     *
     * @var array
     */
    protected static $studlyCache = [];
    
    /**
     * Encoding standard
     *
     * @var string
     */
    protected static $encoding = 'UTF-8';
    
    /**
     * Appends a value to the text using a specified glue
     *
     * @param string $value
     * @param string $text
     * @param string $glue
     *
     * @since 0.1
     * @return string
     */
    public static function append( $value, $text, $glue = '' ) {
        $text = "{$text}{$glue}{$value}";
        
        return $text;
    }
    
    /**
     * Transliterate a UTF-8 value to ASCII.
     *
     * @param string $value
     *
     * @since 0.1
     * @return string
     */
    public static function ascii( $value ) {
        foreach ( static::charsArray() as $key => $val ) {
            $value = str_replace( $val, $key, $value );
        }
        
        return preg_replace( '/[^\x20-\x7E]/u', '', $value );
    }
    
    /**
     * Convert a value to camel case.
     *
     * @param string $value
     *
     * @since 0.1
     * @return string
     */
    public static function camel( $value ) {
        if( isset( static::$camelCache[ $value ] ) ) {
            return static::$camelCache[ $value ];
        }
        
        return static::$camelCache[ $value ] = lcfirst( static::studly( $value ) );
    }
    
    /**
     * Determine if a given string contains a given substring.
     *
     * @param string       $haystack
     * @param string|array $needles
     *
     * @since 0.1
     * @return bool
     */
    public static function contains( $haystack, $needles ) {
        foreach ( (array)$needles as $needle ) {
            if( $needle !== '' && strpos( $haystack, $needle ) !== false ) {
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * Convert unicode of soft hyphen into the equivalent in HTML
     *
     * @param $text
     *
     * @since 0.1
     * @return string
     */
    public static function convertSoftHyphen( $text ) {
        return static::unicodeToHtml( $text, [ "0085" => "&shy;" ] );
    }
    
    /**
     * Determine if a given string ends with a given substring.
     *
     * @param string       $haystack
     * @param string|array $needles
     *
     * @since 0.1
     * @return bool
     */
    public static function endsWith( $haystack, $needles ) {
        foreach ( (array)$needles as $needle ) {
            if( (string)$needle === substr( $haystack, -strlen( $needle ) ) ) {
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * Determine if a given string is empty.
     *
     * @param string $string
     *
     * @since 0.1
     * @return bool
     */
    public static function isEmpty( $string ) {
        return self::length( trim( $string ) ) === 0;
    }
    
    /**
     * Cap a string with a single instance of a given value.
     *
     * @param string $value
     * @param string $cap
     *
     * @since 0.1
     * @return string
     */
    public static function finish( $value, $cap ) {
        $quoted = preg_quote( $cap, '/' );
        
        return preg_replace( '/(?:' . $quoted . ')+$/', '', $value ) . $cap;
    }
    
    /**
     * Determine if a given string matches a given pattern.
     *
     * @param string $pattern
     * @param string $value
     *
     * @since 0.1
     * @return bool
     */
    public static function is( $pattern, $value ) {
        if( $pattern == $value ) {
            return true;
        }
        
        $pattern = preg_quote( $pattern, '#' );
        
        // Asterisks are translated into zero-or-more regular expression wildcards
        // to make it convenient to check if the strings starts with the given
        // pattern such as "library/*", making any string check convenient.
        $pattern = str_replace( '\*', '.*', $pattern );
        
        return (bool)preg_match( '#^' . $pattern . '\z#', $value );
    }
    
    /**
     * Return the length of the given string.
     *
     * @param string $value
     *
     * @since 0.1
     * @return int
     */
    public static function length( $value ) {
        return mb_strlen( $value );
    }
    
    /**
     * Limit the number of characters in a string.
     *
     * @param string $value
     * @param int    $limit
     * @param string $end
     *
     * @since 0.1
     * @return string
     */
    public static function limit( $value, $limit = 100, $end = '...' ) {
        if( mb_strwidth( $value, static::$encoding ) <= $limit ) {
            return $value;
        }
        
        return rtrim( mb_strimwidth( $value, 0, $limit, '', static::$encoding ) ) . $end;
    }
    
    /**
     * Convert the given string to lower-case.
     *
     * @param string $value
     *
     * @since 0.1
     * @return string
     */
    public static function lower( $value ) {
        return mb_strtolower( $value, static::$encoding );
    }
    
    /**
     * Confirm that a given string dose not matches a given pattern.
     *
     * @param string|array $pattern
     * @param string       $value
     *
     * @since 0.1
     * @return bool
     */
    public static function not( $pattern, $value ) {
        return ! static::is( $pattern, $value );
    }
    
    /**
     * Determine if a given string is not empty.
     *
     * @param $string
     *
     * @since 0.1
     * @return bool
     */
    public static function notEmpty( $string ) {
        return ! static::isEmpty( $string );
    }
    
    /**
     * Parse a 'Class@method' style callback into class and method.
     *
     * @param string $callback
     * @param string $default
     *
     * @since 0.1
     * @return array
     */
    public static function parseCallback( $callback, $default ) {
        return static::contains( $callback, '@' ) ? explode( '@', $callback, 2 ) : [ $callback, $default ];
    }
    
    /**
     * prepends a value to the text using a specified glue
     *
     * @param        $value
     * @param string $text
     * @param string $glue
     *
     * @since 0.1
     * @return string
     */
    public static function prepend( $value, $text, $glue = '' ) {
        $text = "{$value}{$glue}{$text}";
        
        return $text;
    }
    
    /**
     * Generate a random string of words
     *
     * @param int $words
     *
     * @since 0.1
     * @return string
     */
    public static function quickLorem( $words = 10 ) {
        $str = [];
        for ( $word = 0; $word < $words; $word++ ) {
            $str[] = static::quickRandom( mt_rand( 3, 15 ) );
        }
        
        return implode( ' ', $str );
    }
    
    /**
     * Generate a "random" alpha-numeric string.
     * Should not be considered sufficient for cryptography, etc.
     *
     * @param int $length
     *
     * @since 0.1
     * @return string
     */
    public static function quickRandom( $length = 16 ) {
        $pool = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
        
        return substr( str_shuffle( str_repeat( $pool, $length ) ), 0, $length );
    }
    
    /**
     * Generate a URL friendly "slug" from a given string.
     *
     * @param string $title
     * @param string $separator
     *
     * @since 0.1
     * @return string
     */
    public static function slug( $title, $separator = '-' ) {
        $title = static::snake( static::ascii( $title ) );
        
        // Convert all dashes/underscores into separator
        $flip = $separator == '-' ? '_' : '-';
        
        $title = preg_replace( '![' . preg_quote( $flip ) . ']+!u', $separator, $title );
        
        // Remove all characters that are not the separator, letters, numbers, or whitespace.
        $title = preg_replace( '![^' . preg_quote( $separator ) . '\pL\pN\s]+!u', '', mb_strtolower( $title ) );
        
        // Replace all separator characters and whitespace by a single separator
        $title = preg_replace( '![' . preg_quote( $separator ) . '\s]+!u', $separator, $title );
        
        return trim( $title, $separator );
    }
    
    /**
     * Convert a string to snake case.
     *
     * @param string $value
     * @param string $delimiter
     *
     * @since 0.1
     * @return string
     */
    public static function snake( $value, $delimiter = '_' ) {
        $key = $value . $delimiter;
        
        if( isset( static::$snakeCache[ $key ] ) ) {
            return static::$snakeCache[ $key ];
        }
        
        if( ! ctype_lower( $value ) ) {
            
            // Convert all dashes/underscores into delimiter
            $flip = $delimiter == '-' ? '_' : '-';
            
            $value = preg_replace( '![' . preg_quote( $flip ) . ']+!u', $delimiter, $value );
            
            $value = preg_replace( '/\s+/', $delimiter, $value );
            $value = strtolower( preg_replace( '/(.)(?=[A-Z])/', '$1' . $delimiter, $value ) );
            $value = preg_replace( '/\\' . $delimiter . $delimiter . '/', $delimiter, $value );
        }
        
        return static::$snakeCache[ $key ] = $value;
    }
    
    /**
     * Determine if a given string starts with a given substring.
     *
     * @param string       $haystack
     * @param string|array $needles
     *
     * @since 0.1
     * @return bool
     */
    public static function startsWith( $haystack, $needles ) {
        foreach ( (array)$needles as $needle ) {
            if( $needle !== '' && strpos( $haystack, $needle ) === 0 ) {
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * Convert a value to studly caps case.
     *
     * @param string $value
     *
     * @since 0.1
     * @return string
     */
    public static function studly( $value ) {
        $key = $value;
        
        if( isset( static::$studlyCache[ $key ] ) ) {
            return static::$studlyCache[ $key ];
        }
        
        $value = ucwords( str_replace( [ '-', '_' ], ' ', $value ) );
        
        return static::$studlyCache[ $key ] = str_replace( ' ', '', $value );
    }
    
    /**
     * Returns the portion of string specified by the start and length parameters.
     *
     * @param string   $string
     * @param int      $start
     * @param int|null $length
     *
     * @since 0.1
     * @return string
     */
    public static function substr( $string, $start, $length = null ) {
        return mb_substr( $string, $start, $length, static::$encoding );
    }
    
    /**
     * Strip unwanted tags from the string
     *
     * @param string $text
     * @param null   $allowed_tags
     *
     * @since 0.1
     * @return string
     */
    public static function stripTags( $text = '', $allowed_tags = null ) {
        return ( is_string( $text ) ) ? trim( strip_tags( $text, $allowed_tags ) ) : $text;
    }
    
    /**
     * Convert the given string to title case.
     *
     * @param string $value
     *
     * @since 0.1
     * @return string
     */
    public static function title( $value ) {
        return mb_convert_case( $value, MB_CASE_TITLE, static::$encoding );
    }
    
    /**
     * Truncate text by a set limit with an optional tolerance
     *
     * @param string $text       Text to be truncated
     * @param int    $max_length The max length of the result string
     * @param array  $options    [optional] options to add extra function when truncating the text
     *
     * @since 0.1
     * @return string
     */
    public static function truncate( $text, $max_length = 400, $options = [] ) {
        
        // Converts a text-array into a string.
        if( is_array( $text ) ) {
            $text = implode( ' ', $text );
        }
        
        // return with no change if string is shorter than $length
        if( static::length( $text ) <= $max_length ) {
            return $text;
        }
        
        $options = array_replace( [
            'tolerance'  => 10,
            'break_word' => false,
            'cap'        => '...'
        ], $options );
        
        $max_tolerance = $max_length + $options[ 'tolerance' ];
        
        // Extract string at tolerated max length and remove the last part if there is a part of a word left where the cut has been made.
        $text = preg_replace('/\W\w+\s*(\W*)$/', '$1', static::substr( $text, 0, $max_tolerance ));
        
        if ( $options[ 'break_word' ] && static::length($text) > $max_length ) {
            $text = static::substr( $text, 0, $max_length );
        }
        
        // Remove specific chars att the end of the text if cap is to be added
        if ( static::notEmpty($options[ 'cap' ]) ) {
            $text = rtrim($text, '.,!?');
        }
        
        return static::finish( $text, $options[ 'cap' ] );
    }
    
    /**
     * Convert an array och unicodes to defined values
     *
     * @param string $text
     * @param array  $codes
     *
     * @since 0.1
     * @return mixed
     */
    public static function unicodeToHtml( $text, $codes = [] ) {
        
        foreach ( $codes as $code => $replace ) {
            $code_match = "/\x{" . $code . "}/u";
            $text       = preg_replace( $code_match, $replace, $text );
        }
        
        return $text;
    }
    
    /**
     * Make a string's first character uppercase.
     *
     * @param string $string
     *
     * @since 0.1
     * @return string
     */
    public static function ucfirst( $string ) {
        return static::upper( static::substr( $string, 0, 1 ) ) . static::substr( $string, 1 );
    }
    
    /**
     * Convert the given string to upper-case.
     *
     * @param string $value
     *
     * @since 0.1
     * @return string
     */
    public static function upper( $value ) {
        return mb_strtoupper( $value, static::$encoding );
    }
    
    /**
     * Limit the number of words in a string.
     *
     * @param string $value
     * @param int    $words
     * @param string $end
     *
     * @since 0.1
     * @return string
     */
    public static function words( $value, $words = 100, $end = '...' ) {
        preg_match( '/^\s*+(?:\S++\s*+){1,' . $words . '}/u', $value, $matches );
        
        if( ! isset( $matches[ 0 ] ) || strlen( $value ) === strlen( $matches[ 0 ] ) ) {
            return $value;
        }
        
        return rtrim( $matches[ 0 ] ) . $end;
    }
    
    /**
     * Count the words in the given string
     *
     * @param $string
     *
     * @since 0.1
     * @return int
     */
    public static function wordCount( $string ) {
        return count( preg_split( '/\s+/', $string ) );
    }
    
    /**
     * Returns the replacements for the ascii method.
     * Note: Adapted from Stringy\Stringy.
     *
     * @see   https://github.com/danielstjules/Stringy/blob/2.1.0/LICENSE.txt
     * @since 0.1
     * @return array
     */
    protected static function charsArray() {
        static $charsArray;
        
        if( isset( $charsArray ) ) {
            return $charsArray;
        }
        
        return $charsArray = [
            'a'    => [
                'à',
                'á',
                'ả',
                'ã',
                'ạ',
                'ă',
                'ắ',
                'ằ',
                'ẳ',
                'ẵ',
                'ặ',
                'â',
                'ấ',
                'ầ',
                'ẩ',
                'ẫ',
                'ậ',
                'ā',
                'ą',
                'å',
                'ä',
                'α',
                'ά',
                'ἀ',
                'ἁ',
                'ἂ',
                'ἃ',
                'ἄ',
                'ἅ',
                'ἆ',
                'ἇ',
                'ᾀ',
                'ᾁ',
                'ᾂ',
                'ᾃ',
                'ᾄ',
                'ᾅ',
                'ᾆ',
                'ᾇ',
                'ὰ',
                'ά',
                'ᾰ',
                'ᾱ',
                'ᾲ',
                'ᾳ',
                'ᾴ',
                'ᾶ',
                'ᾷ',
                'а',
                'أ'
            ],
            'b'    => [ 'б', 'β', 'Ъ', 'Ь', 'ب' ],
            'c'    => [ 'ç', 'ć', 'č', 'ĉ', 'ċ' ],
            'd'    => [ 'ď', 'ð', 'đ', 'ƌ', 'ȡ', 'ɖ', 'ɗ', 'ᵭ', 'ᶁ', 'ᶑ', 'д', 'δ', 'د', 'ض' ],
            'e'    => [
                'é',
                'è',
                'ẻ',
                'ẽ',
                'ẹ',
                'ê',
                'ế',
                'ề',
                'ể',
                'ễ',
                'ệ',
                'ë',
                'ē',
                'ę',
                'ě',
                'ĕ',
                'ė',
                'ε',
                'έ',
                'ἐ',
                'ἑ',
                'ἒ',
                'ἓ',
                'ἔ',
                'ἕ',
                'ὲ',
                'έ',
                'е',
                'ё',
                'э',
                'є',
                'ə'
            ],
            'f'    => [ 'ф', 'φ', 'ف' ],
            'g'    => [ 'ĝ', 'ğ', 'ġ', 'ģ', 'г', 'ґ', 'γ', 'ج' ],
            'h'    => [ 'ĥ', 'ħ', 'η', 'ή', 'ح', 'ه' ],
            'i'    => [
                'í',
                'ì',
                'ỉ',
                'ĩ',
                'ị',
                'î',
                'ï',
                'ī',
                'ĭ',
                'į',
                'ı',
                'ι',
                'ί',
                'ϊ',
                'ΐ',
                'ἰ',
                'ἱ',
                'ἲ',
                'ἳ',
                'ἴ',
                'ἵ',
                'ἶ',
                'ἷ',
                'ὶ',
                'ί',
                'ῐ',
                'ῑ',
                'ῒ',
                'ΐ',
                'ῖ',
                'ῗ',
                'і',
                'ї',
                'и'
            ],
            'j'    => [ 'ĵ', 'ј', 'Ј' ],
            'k'    => [ 'ķ', 'ĸ', 'к', 'κ', 'Ķ', 'ق', 'ك' ],
            'l'    => [ 'ł', 'ľ', 'ĺ', 'ļ', 'ŀ', 'л', 'λ', 'ل' ],
            'm'    => [ 'м', 'μ', 'م' ],
            'n'    => [ 'ñ', 'ń', 'ň', 'ņ', 'ŉ', 'ŋ', 'ν', 'н', 'ن' ],
            'o'    => [
                'ó',
                'ò',
                'ỏ',
                'õ',
                'ọ',
                'ô',
                'ố',
                'ồ',
                'ổ',
                'ỗ',
                'ộ',
                'ơ',
                'ớ',
                'ờ',
                'ở',
                'ỡ',
                'ợ',
                'ø',
                'ö',
                'ō',
                'ő',
                'ŏ',
                'ο',
                'ὀ',
                'ὁ',
                'ὂ',
                'ὃ',
                'ὄ',
                'ὅ',
                'ὸ',
                'ό',
                'о',
                'و',
                'θ'
            ],
            'p'    => [ 'п', 'π' ],
            'r'    => [ 'ŕ', 'ř', 'ŗ', 'р', 'ρ', 'ر' ],
            's'    => [ 'ś', 'š', 'ş', 'с', 'σ', 'ș', 'ς', 'س', 'ص' ],
            't'    => [ 'ť', 'ţ', 'т', 'τ', 'ț', 'ت', 'ط' ],
            'u'    => [ 'ú', 'ù', 'ủ', 'ũ', 'ụ', 'ư', 'ứ', 'ừ', 'ử', 'ữ', 'ự', 'û', 'ū', 'ů', 'ű', 'ŭ', 'ų', 'µ', 'у' ],
            'v'    => [ 'в' ],
            'w'    => [ 'ŵ', 'ω', 'ώ' ],
            'x'    => [ 'χ' ],
            'y'    => [ 'ý', 'ỳ', 'ỷ', 'ỹ', 'ỵ', 'ÿ', 'ŷ', 'й', 'ы', 'υ', 'ϋ', 'ύ', 'ΰ', 'ي' ],
            'z'    => [ 'ź', 'ž', 'ż', 'з', 'ζ', 'ز' ],
            'aa'   => [ 'ع' ],
            'ae'   => [ 'æ' ],
            'ch'   => [ 'ч' ],
            'dj'   => [ 'ђ', 'đ' ],
            'dz'   => [ 'џ' ],
            'gh'   => [ 'غ' ],
            'kh'   => [ 'х', 'خ' ],
            'lj'   => [ 'љ' ],
            'nj'   => [ 'њ' ],
            'oe'   => [ 'œ' ],
            'ps'   => [ 'ψ' ],
            'sh'   => [ 'ш' ],
            'shch' => [ 'щ' ],
            'ss'   => [ 'ß' ],
            'th'   => [ 'þ', 'ث', 'ذ', 'ظ' ],
            'ts'   => [ 'ц' ],
            'ue'   => [ 'ü' ],
            'ya'   => [ 'я' ],
            'yu'   => [ 'ю' ],
            'zh'   => [ 'ж' ],
            '(c)'  => [ '©' ],
            'A'    => [
                'Á',
                'À',
                'Ả',
                'Ã',
                'Ạ',
                'Ă',
                'Ắ',
                'Ằ',
                'Ẳ',
                'Ẵ',
                'Ặ',
                'Â',
                'Ấ',
                'Ầ',
                'Ẩ',
                'Ẫ',
                'Ậ',
                'Å',
                'Ä',
                'Ā',
                'Ą',
                'Α',
                'Ά',
                'Ἀ',
                'Ἁ',
                'Ἂ',
                'Ἃ',
                'Ἄ',
                'Ἅ',
                'Ἆ',
                'Ἇ',
                'ᾈ',
                'ᾉ',
                'ᾊ',
                'ᾋ',
                'ᾌ',
                'ᾍ',
                'ᾎ',
                'ᾏ',
                'Ᾰ',
                'Ᾱ',
                'Ὰ',
                'Ά',
                'ᾼ',
                'А'
            ],
            'B'    => [ 'Б', 'Β' ],
            'C'    => [ 'Ç', 'Ć', 'Č', 'Ĉ', 'Ċ' ],
            'D'    => [ 'Ď', 'Ð', 'Đ', 'Ɖ', 'Ɗ', 'Ƌ', 'ᴅ', 'ᴆ', 'Д', 'Δ' ],
            'E'    => [
                'É',
                'È',
                'Ẻ',
                'Ẽ',
                'Ẹ',
                'Ê',
                'Ế',
                'Ề',
                'Ể',
                'Ễ',
                'Ệ',
                'Ë',
                'Ē',
                'Ę',
                'Ě',
                'Ĕ',
                'Ė',
                'Ε',
                'Έ',
                'Ἐ',
                'Ἑ',
                'Ἒ',
                'Ἓ',
                'Ἔ',
                'Ἕ',
                'Έ',
                'Ὲ',
                'Е',
                'Ё',
                'Э',
                'Є',
                'Ə'
            ],
            'F'    => [ 'Ф', 'Φ' ],
            'G'    => [ 'Ğ', 'Ġ', 'Ģ', 'Г', 'Ґ', 'Γ' ],
            'H'    => [ 'Η', 'Ή' ],
            'I'    => [
                'Í',
                'Ì',
                'Ỉ',
                'Ĩ',
                'Ị',
                'Î',
                'Ï',
                'Ī',
                'Ĭ',
                'Į',
                'İ',
                'Ι',
                'Ί',
                'Ϊ',
                'Ἰ',
                'Ἱ',
                'Ἳ',
                'Ἴ',
                'Ἵ',
                'Ἶ',
                'Ἷ',
                'Ῐ',
                'Ῑ',
                'Ὶ',
                'Ί',
                'И',
                'І',
                'Ї'
            ],
            'K'    => [ 'К', 'Κ' ],
            'L'    => [ 'Ĺ', 'Ł', 'Л', 'Λ', 'Ļ' ],
            'M'    => [ 'М', 'Μ' ],
            'N'    => [ 'Ń', 'Ñ', 'Ň', 'Ņ', 'Ŋ', 'Н', 'Ν' ],
            'O'    => [
                'Ó',
                'Ò',
                'Ỏ',
                'Õ',
                'Ọ',
                'Ô',
                'Ố',
                'Ồ',
                'Ổ',
                'Ỗ',
                'Ộ',
                'Ơ',
                'Ớ',
                'Ờ',
                'Ở',
                'Ỡ',
                'Ợ',
                'Ø',
                'Ö',
                'Ō',
                'Ő',
                'Ŏ',
                'Ο',
                'Ό',
                'Ὀ',
                'Ὁ',
                'Ὂ',
                'Ὃ',
                'Ὄ',
                'Ὅ',
                'Ὸ',
                'Ό',
                'О',
                'Θ',
                'Ө'
            ],
            'P'    => [ 'П', 'Π' ],
            'R'    => [ 'Ř', 'Ŕ', 'Р', 'Ρ' ],
            'S'    => [ 'Ş', 'Ŝ', 'Ș', 'Š', 'Ś', 'С', 'Σ' ],
            'T'    => [ 'Ť', 'Ţ', 'Ŧ', 'Ț', 'Т', 'Τ' ],
            'U'    => [ 'Ú', 'Ù', 'Ủ', 'Ũ', 'Ụ', 'Ư', 'Ứ', 'Ừ', 'Ử', 'Ữ', 'Ự', 'Û', 'Ū', 'Ů', 'Ű', 'Ŭ', 'Ų', 'У' ],
            'V'    => [ 'В' ],
            'W'    => [ 'Ω', 'Ώ' ],
            'X'    => [ 'Χ' ],
            'Y'    => [ 'Ý', 'Ỳ', 'Ỷ', 'Ỹ', 'Ỵ', 'Ÿ', 'Ῠ', 'Ῡ', 'Ὺ', 'Ύ', 'Ы', 'Й', 'Υ', 'Ϋ' ],
            'Z'    => [ 'Ź', 'Ž', 'Ż', 'З', 'Ζ' ],
            'AE'   => [ 'Æ' ],
            'CH'   => [ 'Ч' ],
            'DJ'   => [ 'Ђ' ],
            'DZ'   => [ 'Џ' ],
            'KH'   => [ 'Х' ],
            'LJ'   => [ 'Љ' ],
            'NJ'   => [ 'Њ' ],
            'OE'   => [],
            'PS'   => [ 'Ψ' ],
            'SH'   => [ 'Ш' ],
            'SHCH' => [ 'Щ' ],
            'SS'   => [ 'ẞ' ],
            'TH'   => [ 'Þ' ],
            'TS'   => [ 'Ц' ],
            'UE'   => [ 'Ü' ],
            'YA'   => [ 'Я' ],
            'YU'   => [ 'Ю' ],
            'ZH'   => [ 'Ж' ],
            ' '    => [
                "\xC2\xA0",
                "\xE2\x80\x80",
                "\xE2\x80\x81",
                "\xE2\x80\x82",
                "\xE2\x80\x83",
                "\xE2\x80\x84",
                "\xE2\x80\x85",
                "\xE2\x80\x86",
                "\xE2\x80\x87",
                "\xE2\x80\x88",
                "\xE2\x80\x89",
                "\xE2\x80\x8A",
                "\xE2\x80\xAF",
                "\xE2\x81\x9F",
                "\xE3\x80\x80"
            ],
        ];
    }
}
